package de.fraunhofer.iem.dataprovider.tools.ort.service

import de.fraunhofer.iem.dataprovider.configuration.OpenCodeApiProperties
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.tool.dto.CreateToolDto
import de.fraunhofer.iem.dataprovider.tool.dto.FindingDto
import de.fraunhofer.iem.dataprovider.tool.entity.ToolEntity
import de.fraunhofer.iem.dataprovider.tool.enumeration.ToolType
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import de.fraunhofer.iem.dataprovider.tools.ort.dto.VulnerabilityDto
import de.fraunhofer.iem.dataprovider.tools.ort.json.OrtJson
import de.fraunhofer.iem.dataprovider.tools.ort.json.ResultJson
import de.fraunhofer.iem.dataprovider.utilities.HttpClientWrapper
import io.ktor.client.call.*
import io.ktor.client.request.*
import io.ktor.client.statement.*
import io.ktor.http.*
import org.springframework.stereotype.Service

@Service
class OrtService(
    private val openCodeApiProperties: OpenCodeApiProperties,
    private val toolService: ToolService,
    httpClientWrapper: HttpClientWrapper
) {

    private val logger = getLogger(javaClass)
    private val httpClient = httpClientWrapper.getClient()

    val toolEntity: ToolEntity = getOrCreateToolEntity()

    /**
     * Queries the ORT API and transforms the received JSON objects into
     * VulnerabilityDtos.
     *
     * When the API call fails, or contains malformed/incomplete
     * elements this function returns an empty list.
     */
    suspend fun getOrtResults(repoId: Long): List<VulnerabilityDto> {
        val ortResults = queryOrtApi(repoId)
        logger.info("Got ${ortResults.size} ORT results for $repoId.")
        return ortResults.mapNotNull {
            if (it.cveId != null && it.packageName != null && it.severity != null) {
                VulnerabilityDto(it.cveId, it.packageName, it.severity)
            } else {
                null
            }
        }
    }

    suspend fun getFindings(rawResults: List<VulnerabilityDto>): List<FindingDto> {
        return rawResults.map { result ->
            FindingDto(
                "Vulnerability with CVE ${result.cveIdentifier} " +
                    "found in package ${result.packageName} with severity ${result.severity}"
            )
        }
    }

    private final fun getOrCreateToolEntity(): ToolEntity {
        val createToolDto = CreateToolDto("ORT", ToolType.ORT)
        return toolService.findOrCreateTool(createToolDto)
    }

    private suspend fun queryOrtApi(repoId: Long): List<ResultJson> {
        return try {
            logger.info("Query ORT API for repo with id $repoId")
            val response: HttpResponse = httpClient.get(getToolApiPath(repoId))

            val ortJson = response.body<OrtJson>()
            if (ortJson.code != HttpStatusCode.OK.value) {
                throw Exception("ORT API returned with code ${ortJson.code}")
            }
            logger.info("Query ORT API for repo with id $repoId returned with ${ortJson.code}")
            ortJson.result
        } catch (exception: Exception) {
            logger.error("Query to ORT API failed with exception $exception")
            emptyList()
        }
    }

    private fun getToolApiPath(repoId: Long): String {
        return "${openCodeApiProperties.basePath}$repoId${openCodeApiProperties.ort}"
    }
}
